
"""
    ##########################################
    Simulated studies (``examples.synthetic``)
    ##########################################
    
    This module contains examples of factorization runs. Since the data is artificially generated, 
    this is not a valid test of models applicability to real world situations. It can however
    be used for demonstration of the library. 
    
    Examples are performed on 20 x 30 dense matrix, whose values are drawn from normal 
    distribution with zero mean and variance of one (an absolute of values is taken because of 
    nonnegativity constraint).
    
    Only for the purpose of demonstration in all examples many optional (runtime or algorithm specific) 
    parameters are set. The user could as well run the factorization by providing only the target matrix.
    In that case the defaults would be used. General model parameters are explained in :mod:`nimfa.mf_run`, 
    algorithm specific parameters in Python module implementing the algorithm. Nevertheless for best results, 
    careful choice of parameters is recommended. No tracking is demonstrated here.
    
    .. note:: For most factorizations using artificially generated data is not the intended usage (e. g. SNMNMF is in [Zhang2011]_
              used for identification of the microRNA-gene regulatory networks). Consider this when discussing convergence
              and measurements output. 
        
    To run the examples simply type::
        
        python synthetic.py
        
    or call the module's function::
    
        import nimfa.examples
        nimfa.examples.synthetic.run()
"""

import nimfa
import numpy as np
import scipy.sparse as sp


def __fact_factor(X):
    """
    Return dense factorization factor, so that output is printed nice if factor is sparse.
    
    :param X: Factorization factor.
    :type X: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil, dia or :class:`numpy.matrix`
    """
    return X.todense() if sp.isspmatrix(X) else X


def print_info(fit, idx=None):
    """
    Print to stdout info about the factorization.
    
    :param fit: Fitted factorization model.
    :type fit: :class:`nimfa.models.mf_fit.Mf_fit`
    :param idx: Name of the matrix (coefficient) matrix. Used only in the multiple NMF model. Therefore in factorizations 
                that follow standard or nonsmooth model, this parameter can be omitted. Currently, SNMNMF implements 
                multiple NMF model.
    :type idx: `str` with values 'coef' or 'coef1' (`int` value of 0 or 1, respectively) 
    """
    print("=================================================================================================")
    print("Factorization method:", fit.fit)
    print("Initialization method:", fit.fit.seed)
    print("Basis matrix W: ")
    print(__fact_factor(fit.basis()))
    print("Mixture (Coefficient) matrix H%d: " % (idx if idx != None else 0))
    print(__fact_factor(fit.coef(idx)))
    print("Distance (Euclidean): ", fit.distance(metric='euclidean', idx=idx))
    # We can access actual number of iteration directly through fitted model.
    # fit.fit.n_iter
    print("Actual number of iterations: ", fit.summary(idx)['n_iter'])
    # We can access sparseness measure directly through fitted model.
    # fit.fit.sparseness()
    print("Sparseness basis: %7.4f, Sparseness mixture: %7.4f" % (fit.summary(idx)['sparseness'][0], fit.summary(idx)['sparseness'][1]))
    # We can access explained variance directly through fitted model.
    # fit.fit.evar()
    print("Explained variance: ", fit.summary(idx)['evar'])
    # We can access residual sum of squares directly through fitted model.
    # fit.fit.rss()
    print("Residual sum of squares: ", fit.summary(idx)['rss'])
    # There are many more ... but just cannot print out everything =] and some measures need additional data or more runs
    # e.g. entropy, predict, purity, coph_cor, consensus, select_features, score_features, connectivity
    print("=================================================================================================")


def run_snmnmf(V, V1):
    """
    Run sparse network-regularized multiple NMF. 
    
    :param V: First target matrix to estimate.
    :type V: :class:`numpy.matrix`
    :param V1: Second target matrix to estimate.
    :type V1: :class:`numpy.matrix`
    """
    rank = 10
    snmnmf = nimfa.Snmnmf(V, V1, seed="random_c", rank=rank, max_iter=12,
                          A=sp.csr_matrix((V1.shape[1], V1.shape[1])), B=sp.csr_matrix((V.shape[1], V1.shape[1])),
                          gamma = 0.01, gamma_1 = 0.01, lamb = 0.01, lamb_1 = 0.01)
    fit = snmnmf()
    # print all quality measures concerning first target and mixture matrix in
    # multiple NMF
    print_info(fit, idx=0)
    # print all quality measures concerning second target and mixture matrix
    # in multiple NMF
    print_info(fit, idx=1)


def run_bd(V):
    """
    Run Bayesian decomposition.
    
    :param V: Target matrix to estimate.
    :type V: :class:`numpy.matrix`
    """
    rank = 10
    bd = nimfa.Bd(V, seed="random_c", rank=rank, max_iter=12, alpha=np.zeros((V.shape[0], rank)),
                  beta=np.zeros((rank, V.shape[1])), theta=.0, k=.0, sigma=1., skip=100,
                  stride=1, n_w=np.mat(np.zeros((rank, 1))), n_h=np.mat(np.zeros((rank, 1))),
                  n_sigma=False)
    fit = bd()
    print_info(fit)


def run_bmf(V):
    """
    Run binary matrix factorization.
    
    :param V: Target matrix to estimate.
    :type V: :class:`numpy.matrix`
    """
    rank = 10
    bmf = nimfa.Bmf(V, seed="random_vcol", rank=rank, max_iter=12, initialize_only=True,
                     lambda_w=1.1, lambda_h=1.1)
    fit = bmf()
    print_info(fit)


def run_icm(V):
    """
    Run iterated conditional modes.
    
    :param V: Target matrix to estimate.
    :type V: :class:`numpy.matrix`
    """
    rank = 10
    pnrg = np.random.RandomState()
    icm = nimfa.Icm(V, seed="nndsvd", rank=rank, max_iter=12, iiter=20,
                     alpha=pnrg.randn(V.shape[0], rank),
                     beta=pnrg.randn(rank, V.shape[1]), theta=0., k=0., sigma=1.)
    fit = icm()
    print_info(fit)


def run_lfnmf(V):
    """
    Run local fisher nonnegative matrix factorization.
    
    :param V: Target matrix to estimate.
    :type V: :class:`numpy.matrix`
    """
    rank = 10
    pnrg = np.random.RandomState()
    lfnmf = nimfa.Lfnmf(V, seed=None, W=pnrg.rand(V.shape[0], rank), H=pnrg.rand(rank, V.shape[1]),
                        rank=rank, max_iter=12, alpha=0.01)
    fit = lfnmf()
    print_info(fit)


def run_lsnmf(V):
    """
    Run least squares nonnegative matrix factorization.
    
    :param V: Target matrix to estimate.
    :type V: :class:`numpy.matrix`
    """
    rank = 10
    lsnmf = nimfa.Lsnmf(V, seed="random_vcol", rank=rank, max_iter=12, sub_iter=10,
                        inner_sub_iter=10, beta=0.1, min_residuals=1e-5)
    fit = lsnmf()
    print_info(fit)


def run_nmf(V):
    """
    Run standard nonnegative matrix factorization.
    
    :param V: Target matrix to estimate.
    :type V: :class:`numpy.matrix`
    """
    # Euclidean
    rank = 10
    nmf = nimfa.Nmf(V, seed="random_vcol", rank=rank, max_iter=12, update='euclidean',
                      objective='fro')
    fit = nmf()
    print_info(fit)
    # divergence
    nmf = nimfa.Nmf(V, seed="random_vcol", rank=rank, max_iter=12, initialize_only=True,
                    update='divergence', objective='div')
    fit = nmf()
    print_info(fit)


def run_nsnmf(V):
    """
    Run nonsmooth nonnegative matrix factorization.
    
    :param V: Target matrix to estimate.
    :type V: :class:`numpy.matrix`
    """
    rank = 10
    nsnmf = nimfa.Nsnmf(V, seed="random", rank=rank, max_iter=12, theta=0.5)
    fit = nsnmf()
    print_info(fit)


def run_pmf(V):
    """
    Run probabilistic matrix factorization.
    
    :param V: Target matrix to estimate.
    :type V: :class:`numpy.matrix`
    """
    rank = 10
    pmf = nimfa.Pmf(V, seed="random_vcol", rank=rank, max_iter=12, rel_error=1e-5)
    fit = pmf()
    print_info(fit)


def run_psmf(V):
    """
    Run probabilistic sparse matrix factorization.
    
    :param V: Target matrix to estimate.
    :type V: :class:`numpy.matrix`
    """
    rank = 10
    prng = np.random.RandomState()
    psmf = nimfa.Psmf(V, seed=None, rank=rank, max_iter=12, prior=prng.rand(10))
    fit = psmf()
    print_info(fit)


def run_snmf(V):
    """
    Run sparse nonnegative matrix factorization.
    
    :param V: Target matrix to estimate.
    :type V: :class:`numpy.matrix`
    """
    # SNMF/R
    rank = 10
    snmf = nimfa.Snmf(V, seed="random_c", rank=rank, max_iter=12, version='r', eta=1.,
                       beta=1e-4, i_conv=10, w_min_change=0)
    fit = snmf()
    print_info(fit)
    # SNMF/L
    snmf = nimfa.Snmf(V, seed="random_vcol", rank=rank, max_iter=12, version='l', eta=1.,
                      beta=1e-4, i_conv=10, w_min_change=0)
    fit = snmf()
    print_info(fit)


def run_sepnmf(V):
    """
    Run standard nonnegative matrix factorization.

    :param V: Target matrix to estimate.
    :type V: :class:`numpy.matrix`
    """
    # Euclidean
    rank = 10

    sepnmf = nimfa.SepNmf(V, rank=rank, selection='spa')
    fit = sepnmf()
    print_info(fit)

    sepnmf = nimfa.SepNmf(V, rank=rank, selection='xray')
    fit = sepnmf()
    print_info(fit)

    sepnmf = nimfa.SepNmf(V, rank=rank, compression='qr', selection='xray')
    fit = sepnmf()
    print_info(fit)

    sepnmf = nimfa.SepNmf(V, rank=rank, compression='qr', selection='spa')
    fit = sepnmf()
    print_info(fit)

    sepnmf = nimfa.SepNmf(V, rank=rank, compression='structured',
                          selection='xray')
    fit = sepnmf()
    print_info(fit)

    sepnmf = nimfa.SepNmf(V, rank=rank, compression='structured',
                          selection='xray')
    fit = sepnmf()
    print_info(fit)

    sepnmf = nimfa.SepNmf(V, rank=rank, compression='count_gauss',
                          selection='xray')
    fit = sepnmf()
    print_info(fit)

    sepnmf = nimfa.SepNmf(V, rank=rank, compression='count_gauss',
                          selection='spa')
    fit = sepnmf()
    print_info(fit)

    sepnmf = nimfa.SepNmf(V, rank=rank, compression='count_gauss',
                          selection='none')
    fit = sepnmf()
    print_info(fit)


def run(V=None, V1=None):
    """
    Run examples.
    
    :param V: Target matrix to estimate.
    :type V: :class:`numpy.matrix`
    :param V1: (Second) Target matrix to estimate used in multiple NMF (e. g. SNMNMF).
    :type V1: :class:`numpy.matrix`
    """
    if V is None or V1 is None:
        prng = np.random.RandomState(42)
        # construct target matrices
        V = prng.rand(20, 30)
        V1 = prng.rand(20, 25)
    run_snmnmf(V, V1)
    run_bd(V)
    run_bmf(V)
    run_icm(V)
    run_lfnmf(V)
    run_lsnmf(V)
    run_nmf(V)
    run_nsnmf(V)
    run_pmf(V)
    run_psmf(V)
    run_snmf(V)
    run_sepnmf(V)


if __name__ == "__main__":
    prng = np.random.RandomState(42)
    V = prng.rand(20, 30)
    V1 = prng.rand(20, 25)
    run(V, V1)



import numpy as np

import nimfa

V = np.random.rand(40, 100)
nmf = nimfa.Nmf(V, rank=10, seed="random_vcol", max_iter=200, update='euclidean',
                objective='conn', conn_change=40)
nmf_fit = nmf()



import nimfa

V = nimfa.examples.medulloblastoma.read(normalize=True)

lsnmf = nimfa.Lsnmf(V, seed='random_vcol', rank=50, max_iter=100)
lsnmf_fit = lsnmf()

print('Rss: %5.4f' % lsnmf_fit.fit.rss())
print('Evar: %5.4f' % lsnmf_fit.fit.evar())
print('K-L divergence: %5.4f' % lsnmf_fit.distance(metric='kl'))
print('Sparseness, W: %5.4f, H: %5.4f' % lsnmf_fit.fit.sparseness())




"""    
    This package contains implementations of matrix factorization algorithms. 
"""

from .bd import *
from .icm import *
from .lfnmf import *
from .lsnmf import *
from .nmf import *
from .nsnmf import *
from .pmf import *
from .psmf import *
from .snmf import *
from .bmf import *
from .snmnmf import *
from .pmfcc import *
from .sepnmf import *

methods = {"bd": Bd,
           "icm": Icm,
           "lfnmf": Lfnmf,
           "lsnmf": Lsnmf,
           "nmf": Nmf,
           "nsnmf": Nsnmf,
           "pmf": Pmf,
           "psmf": Psmf,
           "snmf": Snmf,
           "bmf": Bmf,
           "snmnmf": Snmnmf,
           "pmfcc": Pmfcc,
           "sepnmf": SepNmf,
           "none": None}



import numpy as np

import nimfa

V = np.random.rand(23, 200)

# Factorization will be run 3 times (n_run) and factors will be tracked for computing
# cophenetic correlation. Note increased time and space complexity
bmf = nimfa.Bmf(V, max_iter=10, rank=30, n_run=3, track_factor=True)
bmf_fit = bmf()

print('K-L divergence: %5.3f' % bmf_fit.distance(metric='kl'))

sm = bmf_fit.summary()
print('Rss: %5.3f' % sm['rss'])
print('Evar: %5.3f' % sm['evar'])
print('Iterations: %d' % sm['n_iter'])
print('Cophenetic correlation: %5.3f' % sm['cophenetic'])




"""
###################################
Nndsvd (``methods.seeding.nndsvd``)
###################################

Nonnegative Double Singular Value Decomposition (NNDSVD) [Boutsidis2007]_ is a
new method designed to enhance the initialization stage of the nonnegative
matrix factorization. The basic algorithm contains no randomization and is
based on two SVD processes, one approximating the data matrix, the other
approximating positive sections of the resulting partial SVD factors utilizing
an algebraic property of unit rank matrices.

NNDSVD is well suited to initialize NMF algorithms with sparse factors.
Numerical examples suggest that NNDSVD leads to rapid reduction of the
approximation error of many NMF algorithms. By setting algorithm options
:param:`flag` dense factors can be generated.
"""

from nimfa.utils.utils import *
from nimfa.utils.linalg import *

__all__ = ['Nndsvd']


class Nndsvd(object):

    def __init__(self):
        self.name = "nndsvd"

    def initialize(self, V, rank, options):
        """
        Return initialized basis and mixture matrix. 
        
        Initialized matrices are sparse :class:`scipy.sparse.csr_matrix` if
        NNDSVD variant is specified by the :param:`flag` option,
        else matrices are :class:`numpy.matrix`.
        
        :param V: Target matrix, the matrix for MF method to estimate.
                Data instances to be clustered.
        :type V: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil,
                dia or :class:`numpy.matrix`
        :param rank: Factorization rank.
        :type rank: `int`
        :param options: Specify the algorithm and model specific options (e.g. initialization of
                extra matrix factor, seeding parameters).

                Option ``flag`` indicates the variant of the NNDSVD algorithm. It can take value from:

                    * 0 -- NNDSVD,
                    * 1 -- NNDSVDa (fill in the zero elements with the average),
                    * 2 -- NNDSVDar (fill in the zero elements with random values in the space [0:average/100]).

                Default is NNDSVD. Because of the nature of NNDSVDa and NNDSVDar, when the
                target matrix is sparse, only NNDSVD is possible and `flag` is ignored
                (NNDSVDa and NNDSVDar eliminate zero elements, therefore the matrix is
                not sparse anymore).
        :type options: `dict`
        """
        self.rank = rank
        self.flag = options.get('flag', 0)
        if negative(V):
            raise MFError("The input matrix contains negative elements.")
        U, S, E = svd(V)
        E = E.T
        if sp.isspmatrix(U):
            return self.init_sparse(V, U, S, E)
        self.W = np.mat(np.zeros((V.shape[0], self.rank)))
        self.H = np.mat(np.zeros((self.rank, V.shape[1])))
        # choose the first singular triplet to be nonnegative
        S = np.diagonal(S)
        self.W[:, 0] = sqrt(S[0]) * abs(U[:, 0])
        self.H[0, :] = sqrt(S[0]) * abs(E[:, 0].T)
        # second svd for the other factors
        for i in range(1, self.rank):
            uu = U[:, i]
            vv = E[:, i]
            uup = self._pos(uu)
            uun = self._neg(uu)
            vvp = self._pos(vv)
            vvn = self._neg(vv)
            n_uup = norm(uup, 2)
            n_vvp = norm(vvp, 2)
            n_uun = norm(uun, 2)
            n_vvn = norm(vvn, 2)
            termp = n_uup * n_vvp
            termn = n_uun * n_vvn
            if (termp >= termn):
                self.W[:, i] = sqrt(S[i] * termp) / n_uup * uup
                self.H[i, :] = sqrt(S[i] * termp) / n_vvp * vvp.T
            else:
                self.W[:, i] = sqrt(S[i] * termn) / n_uun * uun
                self.H[i, :] = sqrt(S[i] * termn) / n_vvn * vvn.T
        self.W[self.W < 1e-11] = 0
        self.H[self.H < 1e-11] = 0
        # NNDSVD
        if self.flag == 0:
            if sp.isspmatrix(V):
                return sp.lil_matrix(self.W).tocsr(), sp.lil_matrix(self.H).tocsr()
            else:
                return self.W, self.H
        # NNDSVDa
        if self.flag == 1:
            avg = V.mean()
            self.W[self.W == 0] = avg
            self.H[self.H == 0] = avg
        # NNDSVDar
        if self.flag == 2:
            avg = V.mean()
            n1 = len(self.W[self.W == 0])
            n2 = len(self.H[self.H == 0])
            self.W[self.W == 0] = avg * np.random.uniform(n1, 1) / 100
            self.H[self.H == 0] = avg * np.random.uniform(n2, 1) / 100
        return self.W, self.H

    def init_sparse(self, V, U, S, E):
        """
        Continue the NNDSVD initialization of sparse target matrix.
        
        :param V: Target matrix
        :type V: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil, dia
        :param U: Left singular vectors.
        :type U: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil, dia
        :param E: Right singular vectors.
        :type E: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil, dia
        :param S: Singular values.
        :type S: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil, dia
        """
        # LIL sparse format is convenient for construction
        self.W = sp.lil_matrix((V.shape[0], self.rank))
        self.H = sp.lil_matrix((self.rank, V.shape[1]))
        # scipy.sparse.linalg ARPACK does not allow computation of rank(V) eigenvectors
        # fill the missing columns/rows with random values
        prng = np.random.RandomState()
        S = [S[i, i] for i in range(np.min([S.shape[0], S.shape[1]]))]
        S += [prng.rand() for _ in range(self.rank - len(S))]
        U = U.tolil()
        E = E.tolil()
        temp_U = sp.lil_matrix((V.shape[0], min(V.shape[0], V.shape[1])))
        temp_E = sp.lil_matrix((V.shape[1], min(V.shape[0], V.shape[1])))
        if temp_U.shape != U.shape:
            temp_U[:, :U.shape[1]] = U
            temp_U[:, U.shape[1]:] = abs(
                sp.rand(U.shape[0], temp_U.shape[1] - U.shape[1], density=0.8, format='lil'))
        if temp_E.shape != E.shape:
            temp_E[:E.shape[0], :] = E
            temp_E[E.shape[0]:, :] = abs(
                sp.rand(temp_E.shape[0] - E.shape[0], E.shape[1], density=0.8, format='lil'))
        # choose the first singular triplet to be nonnegative
        self.W[:, 0] = sqrt(S[0]) * abs(U[:, 0])
        self.H[0, :] = sqrt(S[0]) * abs(E[:, 0].T)
        eps = np.finfo(V.data.dtype).eps if not 'int' in str(
            V.data.dtype) else 0
        # second svd for the other factors
        for i in range(1, self.rank):
            uu = U[:, i]
            vv = E[:, i]
            uup = self._pos(uu)
            uun = self._neg(uu)
            vvp = self._pos(vv)
            vvn = self._neg(vv)
            n_uup = norm(uup, 2) + eps
            n_vvp = norm(vvp, 2) + eps
            n_uun = norm(uun, 2) + eps
            n_vvn = norm(vvn, 2) + eps
            termp = n_uup * n_vvp
            termn = n_uun * n_vvn
            if (termp >= termn):
                self.W[:, i] = sqrt(S[i] * termp) / n_uup * uup
                self.H[i, :] = sqrt(S[i] * termp) / n_vvp * vvp.T
            else:
                self.W[:, i] = sqrt(S[i] * termn) / n_uun * uun
                self.H[i, :] = sqrt(S[i] * termn) / n_vvn * vvn.T
        # CSR sparse format is convenient for fast arithmetic and matrix vector
        # operations
        return self.W, self.H

    def _pos(self, X):
        """Return positive section of matrix or vector."""
        if sp.isspmatrix(X):
            return multiply(sop(X, 0, ge), X)
        else:
            return multiply(X >= 0, X)

    def _neg(self, X):
        """Return negative section of matrix or vector."""
        if sp.isspmatrix(X):
            return multiply(sop(X, 0, le), - X)
        else:
            return multiply(X < 0, -X)

    def __repr__(self):
        return "nndsvd.Nndsvd()"

    def __str__(self):
        return self.name




"""
    ##########################
    Nmf_ns (``models.nmf_ns``)
    ##########################
"""

from .nmf import *


class Nmf_ns(Nmf):
    """
    Implementation of the alternative model to manage factorizations that follow
    nonstandard NMF model. This modification is required by the Nonsmooth NMF
    algorithm (NSNMF) [Montano2006]_. The Nonsmooth NMF algorithm is a modification
    of the standard divergence based NMF methods. By introducing a smoothing matrix
    it is aimed to achieve global sparseness.
     
    It is the underlying model of matrix factorization and provides structure of
    modified standard NMF model.
    
    .. attribute:: W
        
        Basis matrix -- the first matrix factor in the nonsmooth NMF model
        
    .. attribute:: H
    
        Mixture matrix -- the third matrix factor in the nonsmooth NMF model
        
    .. attribute:: S
    
        Smoothing matrix -- the middle matrix factor (V = WSH) in the nonsmooth NMF model
        
    The interpretation of the basis and mixture matrix is such as in the standard NMF model.
    The smoothing matrix is an extra square matrix whose entries depends on smoothing
    parameter theta which can be specified as algorithm specific model option. For detailed
    explanation of the NSNMF algorithm see :mod:`methods.factorization.nsnmf`.
    """
    def __init__(self, params):
        """
        Construct factorization model that manages nonsmooth NMF models.
        
        :param params: MF runtime and algorithm parameters and options. For detailed
           explanation of the general model parameters see :mod:`mf_run`. For
           algorithm specific model options see documentation of chosen
           factorization method.
        :type params: `dict`
        """
        self.model_name = "ns"
        self.V1 = None
        self.H1 = None
        Nmf.__init__(self, params)
        if sp.isspmatrix(self.V) and (self.V.data < 0).any() or not sp.isspmatrix(self.V) and (self.V < 0).any():
            raise utils.MFError("The input matrix contains negative elements.")

    def basis(self):
        """Return the matrix of basis vectors."""
        return self.W

    def target(self, idx=None):
        """
        Return the target matrix to estimate.
        
        :param idx: Used in the multiple NMF model. In nonsmooth NMF ``idx`` is always None.
        :type idx: None
        """
        return self.V

    def coef(self, idx=None):
        """
        Return the matrix of mixture coefficients.
        
        :param idx: Used in the multiple NMF model. In nonsmooth NMF ``idx`` is always None.
        :type idx: None
        """
        return self.H

    def smoothing(self):
        """Return the smoothing matrix."""
        return self.S

    def fitted(self, idx=None):
        """
        Compute the estimated target matrix according to the nonsmooth NMF algorithm model.
        
        :param idx: Used in the multiple NMF model. In nonsmooth NMF ``idx`` is always None.
        :type idx: None
        """
        return dot(dot(self.W, self.S), self.H)

    def distance(self, metric='euclidean', idx=None):
        """
        Return the loss function value.
        
        :param distance: Specify distance metric to be used. Possible are Euclidean and
           Kullback-Leibler (KL) divergence. Strictly, KL is not a metric.
        :type distance: `str` with values 'euclidean' or 'kl'

        :param idx: Used in the multiple NMF model. In nonsmooth NMF ``idx`` is always None.
        :type idx: None
        """
        if metric.lower() == 'euclidean':
            R = self.V - dot(dot(self.W, self.S), self.H)
            return power(R, 2).sum()
        elif metric.lower() == 'kl':
            Va = dot(dot(self.W, self.S), self.H)
            return (multiply(self.V, sop(elop(self.V, Va, div), op=np.log)) - self.V + Va).sum()
        else:
            raise utils.MFError("Unknown distance metric.")

    def residuals(self, idx=None):
        """
        Return residuals matrix between the target matrix and its nonsmooth NMF estimate.
        
        :param idx: Used in the multiple NMF model. In nonsmooth NMF ``idx`` is always None.
        :type idx: None
        """
        return self.V - dot(dot(self.W, self.S), self.H)




"""
#########################################
SepNmf (``methods.factorization.sepnmf``)
#########################################

**Separable Nonnegative Matrix Factorization (SepNMF)** [Damle2014]_, [Benson2014]_,
[Kumar2013]_, [Gillis2014]_, [Tepper2015]_, [Kapralov2016]_

Separable NMF was introduced by Donoho and Stodden (2003) and polynomial time algorithms were given by Arora et al 2012.
Other algorithms such as XRAY [Kumar2013]_, SPA [Gillis2014]_ and more recently SC
[Tepper2015]_ and CG [Kapralov2016]_ have been proposed.

SepNMF can be used for problems which satisfy the ``pure-pixel'' assumption which occurs in
hyper-spectral imaging and document analysis settings.

"""
try:
    from future_builtins import zip
except ImportError: # not 2.6+ or is 3.x
    try:
        from itertools import izip as zip # < 2.5 or 3.x
    except ImportError:
        pass
import collections
import scipy.optimize as optimize
import scipy.sparse as sp
from nimfa.models import *
from nimfa.utils.linalg import *
from nimfa.utils.utils import *

__all__ = ['SepNmf']


class SepNmf(nmf_std.Nmf_std):
    """
    :param V: The target matrix to estimate.
    :type V: Instance of the :class:`scipy.sparse` sparse matrices types,
       :class:`numpy.ndarray`, :class:`numpy.matrix` or tuple of instances of
       the latter classes.

    :param rank: The factorization rank to achieve. Default is 30.
    :type rank: `int`

    :param n_run: It specifies the number of runs of the algorithm. Default is
       1. If multiple runs are performed, fitted factorization model with the
       lowest objective function value is retained.
    :type n_run: `int`

    :param callback: Pass a callback function that is called after each run when
       performing multiple runs. This is useful if one wants to save summary
       measures or process the result before it gets discarded. The callback
       function is called with only one argument :class:`models.mf_fit.Mf_fit` that
       contains the fitted model. Default is None.
    :type callback: `function`

    :param callback_init: Pass a callback function that is called after each
       initialization of the matrix factors. In case of multiple runs the function
       is called before each run (more precisely after initialization and before
       the factorization of each run). In case of single run, the passed callback
       function is called after the only initialization of the matrix factors.
       This is useful if one wants to obtain the initialized matrix factors for
       further analysis or additional info about initialized factorization model.
       The callback function is called with only one argument
       :class:`models.mf_fit.Mf_fit` that (among others) contains also initialized
       matrix factors. Default is None.
    :type callback_init: `function`

    :param track_factor: When :param:`track_factor` is specified, the fitted
        factorization model is tracked during multiple runs of the algorithm. This
        option is taken into account only when multiple runs are executed
        (:param:`n_run` > 1). From each run of the factorization all matrix factors
        are retained, which can be very space consuming. If space is the problem
        setting the callback function with :param:`callback` is advised which is
        executed after each run. Tracking is useful for performing some quality or
        performance measures (e.g. cophenetic correlation, consensus matrix,
        dispersion). By default fitted model is not tracked.
    :type track_factor: `bool`

    :param track_error: Tracking the residuals error. Only the residuals from
        each iteration of the factorization are retained. Error tracking is not
        space consuming. By default residuals are not tracked and only the final
        residuals are saved. It can be used for plotting the trajectory of the
        residuals.
    :type track_error: `bool`
    """
    def __init__(self, V, rank=30, test_conv=None, n_run=1, callback=None,
                 callback_init=None, track_factor=False, track_error=False,
                 compression=None, selection='spa', **options):
        params = vars()
        del params['self']
        params.update({
            'name': 'sepnmf',
            'W': None,
            'H': None,
            'seed': None,
            'aseeds': 'none',
        })
        super(SepNmf, self).__init__(params)

        if compression != 'count_gauss' and selection == 'none':
            raise MFError('Trivial selection only works with count_gauss '
                          'compression')
        if self.compression is not None:
            self.compression = getattr(self, self.compression)
        self.selection = getattr(self, self.selection)
        if self.track_factor and self.n_run > 1 or self.track_error:
            self.tracker = mf_track.Mf_track()
        else:
            self.tracker = None

    def factorize(self):
        """
        Compute matrix factorization.

        Return fitted factorization model.
        """
        for run in range(self.n_run):
            self.H = zeros_for(self.V, (self.rank, self.V.shape[1]))
            if self.compression:
                V = self.compression()
            else:
                V = self.V
            cols = self.selection(V)
            self.H, obj = nnls(V, V[:, cols])
            self.W = self.V[:, cols]

            if self.callback:
                self.final_obj = obj
                self.n_iter = iter
                mffit = mf_fit.Mf_fit(self)
                self.callback(mffit)
            if self.track_factor:
                self.tracker.track_factor(run, W=self.W, H=self.H,
                                          final_obj=obj, n_iter=0)
            # if multiple runs are performed, fitted factorization model with
            # the lowest objective function value is retained
            if run == 0 or obj <= best_obj:
                best_obj = obj
                self.n_iter = 0
                self.final_obj = obj
                mffit = mf_fit.Mf_fit(copy.deepcopy(self))

        mffit.fit.tracker = self.tracker
        return mffit

    def __str__(self):
        str_format = '{0} - compression: {1} - selection: {2}'
        comp = self.compression.__name__ if self.compression else 'none'
        return str_format.format(self.name, comp, self.selection.__name__)

    def __repr__(self):
        return self.name

    def count_gauss(self, oversampling_factor=5):
        """
        Project the columns of the matrix V into the lower dimension
        new_dim using count sketch + gaussian algorithm
        """
        old_dim = self.V.shape[0]

        # ksq = new_dim * new_dim
        ksq = self.rank * oversampling_factor  # was not converging for scree plots of damle/sun

        R = np.random.randint(ksq, size=old_dim)
        C = np.arange(old_dim)
        D = np.random.choice([-1, 1], size=old_dim)
        S = scipy.sparse.csr_matrix((D, (R, C)), shape=(ksq, old_dim))

        G = np.random.randn(self.rank, ksq)
        M_red = dot(G, dot(S, self.V))
        return M_red.todense() / np.sqrt(self.rank)

    def structured(self, n_power_iter=0, oversampling=10, min_comp=20):
        """
        Structured compression algorithm
        """
        n = self.V.shape[1]
        comp_level = min(max(min_comp, self.rank + oversampling), n)
        omega = np.random.standard_normal(size=(n, comp_level))

        mat_h = self.V.dot(omega)
        for _ in range(n_power_iter):
            mat_h = self.V.dot(self.V.T.dot(mat_h))
        q, _ = np.linalg.qr(mat_h)
        return q.T.dot(self.V)

    def qr(self):
        """
        QR compression algorithm
        """
        q, _ = np.linalg.qr(self.V)
        return q.T.dot(self.V)

    def spa(self, V):
        """
        Successive projection algorithm (SPA) algorithm for extreme
        column selection in separable NMF.
        :param V: The data matrix.
        :type V: Instance of the :class:`scipy.sparse` sparse matrices
        types, :class:`numpy.ndarray`, or :class:`numpy.matrix`.
        :return: The indices of the columns chosen by SPA.
        :rtype: List of ints.
        """
        colnorms = norm_axis(V, p=1, axis=0)
        x = elop(V, colnorms, div)
        cols = []
        m, n = x.shape
        for _ in range(self.rank):
            col_norms = norm_axis(x, axis=0)
            col_norms[0, cols] = -1
            _, col_ind = argmax(col_norms)
            cols.append(col_ind)
            col = x[:, col_ind]  # col is a column vector
            x = dot(np.eye(m) - dot(col, col.T) / col_norms[0, col_ind], x)
        return cols

    def xray(self, V):
        """
        X-ray algorithm for extreme column selection in separable NMF.
        :param V: The data matrix.
        :type V: Instance of the :class:`scipy.sparse` sparse matrices
        types, :class:`numpy.ndarray`, or :class:`numpy.matrix`.
        :return: The indices of the columns chosen by X-ray.
        :rtype: List of ints.
        """
        cols = []
        R = V
        while len(cols) < self.rank:
            # Loop until we choose a column that has not been selected.
            while True:
                p = np.random.random((1, V.shape[0]))
                scores = norm_axis(dot(R.T, V), axis=0)
                scores = elop(scores, dot(p, V), div)
                scores[0, cols] = -1  # IMPORTANT
                best_col = np.argmax(scores)
                if best_col in cols:
                    # Re-try
                    continue
                else:
                    cols.append(best_col)
                    H, _ = nnls(V, V[:, cols])
                    R = V - dot(V[:, cols], H)
                    break
        return cols

    def none(self, V):
        """
        Trivial algorithm for extreme column selection in separable NMF.
        :param V: The data matrix.
        :type V: Instance of the :class:`scipy.sparse` sparse matrices
        types, :class:`numpy.ndarray`, or :class:`numpy.matrix`.
        :return: The indices of the columns chosen by X-ray.
        :rtype: List of ints.
        """
        idx = collections.Counter()
        idx.update(argmax(V, axis=0)[1].tolist()[0])
        idx.update(argmin(V, axis=0)[1].tolist()[0])
        return next(zip(*idx.most_common(self.rank)))


def objective(V, W, H):
    """
    Compute squared Frobenius norm of a target matrix and its NMF
    estimate.
    """
    R = V - dot(W, H)
    return multiply(R, R).sum()


def nnls(V, W):
    """
    Compute H, the coefficient matrix, by nonnegative least squares
    to minimize the Frobenius norm.  Given the data matrix V and the
    basis matrix W, H is
    .. math:: \arg\min_{Y \ge 0} \| V - W H \|_F.
    :param V: The data matrix.
    :type V: numpy.ndarray
    :param W: The data matrix.
    :type W: numpy.ndarray
    :return: The matrix H and the relative residual.
    """
    ncols = V.shape[1]
    H = []
    total_res = 0
    for i in range(ncols):
        b = np.squeeze(np.array(V[:, i]))
        sol, res = optimize.nnls(np.array(W), b)
        H.append(sol[:, np.newaxis])
        total_res += res ** 2
    return np.asmatrix(hstack(H)), total_res


def zeros_for(mat, size):
    if sp.isspmatrix(mat):
        format_dict = {'bsr': sp.bsr_matrix, 'coo': sp.coo_matrix,
                       'csc': sp.csc_matrix, 'csr': sp.csr_matrix,
                       'dia': sp.dia_matrix, 'dok': sp.dok_matrix,
                       'lil': sp.lil_matrix}
        cls = format_dict[format]
        return cls(size)
    else:
        return np.zeros(size)


def norm_axis(X, p=None, axis=None):
    """
    Compute entry-wise norms (! not induced/operator norms).

    :param X: The input matrix.
    :type X: :class:`scipy.sparse` of format csr, csc, coo, bsr, dok, lil, dia
    or :class:`numpy.matrix`
    :param p: Order of the norm.
    :type p: `str` or `float`
    """
    assert 1 in X.shape or p != 2 or axis is not None,\
        "Computing entry-wise norms only."
    if sp.isspmatrix(X):
        return sla.norm(X, ord=p, axis=axis)
    else:
        n = nla.norm(np.mat(X), ord=p, axis=axis)
        if axis is None:
            return n
        else:
            return np.asmatrix(n)
